/*
 * ad7766.c - sport ad7766 driver
 * Uses the bfin_sport device driver, which uses dma descriptor list 
 * to write sport data into a circular buffer queue.
 * Uses the bflin_timer device driver to clock the adc.
 * ------
 * Embedded Ocean, March 2014
 */

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <cycle_count.h>
#include <asm/bfin_sport.h>
#include <asm/bfin_timer.h>
#include <fract.h>

#include "ad7766.h"
#include "gpio.h"

#define GPIO_VREF GPIO_PG7

#define EXT_RSCLK

extern int debug_level;
extern int verbose;

extern FILE *dbg_fp;
extern FILE *err_fp;

static int mclk_fd;
static int rsclk_fd;

// These sampling rates have bee tested.
// ad7766_open: fs=125000, mclk=1000000, rsclk=5000000, sclk=125000000
// ad7766_open: fs=93750, mclk=750000, rsclk=3750000, sclk=125000000
// ad7766_open: fs=62500, mclk=500000, rsclk=2500000, sclk=125000000

adc_t *ad7766_open(char *device, int fs)
{
   int fd, n;
   adc_t *adc;
  
   struct sport_config config;

   if((fs != 125000) && (fs != 93750) && (fs != 62500)) {
      fprintf(err_fp, "ad7766_open: fs = %d not available, use 125k, 93.75k or 62.5k.\n", fs); 
      return(NULL);
   } 

   // open sport device
   fd = open(device, O_RDONLY, 0);
   if (fd < 0) {
      fprintf(err_fp, "ad7766: Failed to open %s\n", device );
      return(NULL);
   }

   // sport recieve control
   memset(&config, 0, sizeof (struct sport_config));

   // use these settings with timer1 as rclk
   config.rfsr = 1;   // frame sync required
   config.larfs = 1;  // late frame sync
   config.irfs = 0;   // frame sync is generated externeally - by adc
   config.lrfs = 0;   // active high frame sync
   config.irclk = 0;  // serial clk generted external - by timer1
   config.rckfe = 0;  // sample data with rising edge of rclk
   config.word_len = 24;

   /* Configure sport controller by ioctl */
   if (ioctl(fd, SPORT_IOC_CONFIG, &config) < 0) {
      fprintf(err_fp, "ad7766_open: failed to config sport\n");
      close(fd);
      return(NULL);
   }

   // turn ADC VREF ON
   gpio_export(GPIO_VREF);
   gpio_dir_out(GPIO_VREF);
   gpio_write(GPIO_VREF, 0);
   usleep(1000);
   gpio_write(GPIO_VREF, 1);
   usleep(1000);

   // reset adc
   //gpio_adc_reset = gpio_open(GPIO_PG3, 'O');
   //gpio_write(gpio_adc_reset, '0');
   //gpio_write(gpio_adc_reset, '1');

   // alloc adc object
   adc = (adc_t *)malloc(sizeof(adc_t));
   if(adc == NULL) {
      fprintf(err_fp, "ad7766_open: error malloc adc\n");
      return(NULL);
   }
   //memset(adc, 0, sizeof (adc_t));
   strncpy(adc->device, device, 32);
   adc->fd = fd;
   adc->fs = fs;

   // set initial/default adc variables 
   adc->nsize = 2; // 2 bytes per sample
   adc->shift = 8; // bit shift
   adc->nbufs = 3; // number of dma buffers, should be > 2
   for (n=0; n<(adc->nbufs); n++) {
      adc->buffer[n] = NULL;
   }
   adc->gain = 0;
   adc->auto_scale = 0; // not used
   adc->clip = 0;  // not used
   adc->nchans = 1;  // not used

   return(adc);
}

/*
 * Allocate adc data buffers for dma data transfers.
 * Buffers can be allocated from kernel managed memory (using malloc) or non-kernel memory.
 * If the compiler directive ADC_USE_PMEM is defined, then a physical (non-kernel) memory region is used for the buffers.
 * If ADC_USE_PMEM is defined, then you need to define BFIN_PMEM_START as the start of the physical mem region.
 * Kernel memory is set in the uboot bootargs.  
 * For example, adding the bootargs mem=32M max_mem=64M$# will tell the kernel to use 32M of the total 64M. 
 * This leaves 32M free as non-kernel mem.
 * Now use #define BFIN_PMEM_START 0x2000000 sets the physical mem start at 32M.
 * If BFIN_PMEM_START is within the kernel memory, then the behavior is unpredictable.
 * 
 */
int ad7766_alloc_dma_buffers(adc_t *adc, float adc_duration, int adc_nsize)
{
   size_t nbytes;
   unsigned long blksize;
   int nblks, n;
  
   blksize = 4096;

   adc->nsize = adc_nsize;  // number bytes per sample
   adc->nsamps = (int)(adc_duration * (float)adc->fs);
   
   // number of data blocks in each adc buffer
   nblks = (adc->nsamps * adc->nsize) / (int)blksize; 
   adc->nsamps = (int)(nblks * blksize) / adc->nsize;  // nsamps per block
   adc_duration = (float)adc->nsamps / (float)adc->fs;

   if(verbose > 1) 
      fprintf(dbg_fp, "ad7766_alloc_dma_buffers: adc buffer duration = %f\n", adc_duration);
   
   if(debug_level > 1) 
      fprintf(dbg_fp, "ad7766_alloc_dma_buffers: adc nblks = %d, nsamps = %d\n", nblks, adc->nsamps);

   // assign data buffers using malloc
   nbytes = 4*adc->nsamps; // always use 4 byte words for adc read buffer
    
   if(debug_level > 0) 
	   fprintf(dbg_fp,"ad7766_alloc_dma_buffers: %ld bytes required for adc buffers\n", (nbytes * adc->nbufs));
   

#ifdef ADC_USE_PMEM
   // assign data buffers using available phys mem
   void *pmem;  // non-kernel physical memory pointer
   size_t poff;
   pmem = (void *)BFIN_PMEM_START;
   poff = 0;
   for(n=0; n<(adc->nbufs); n++) {
      adc->buffer[n] = (char *)(pmem+poff); // bit words
      poff += (nbytes/4096+1)*4096; // increment on even page boundaries 
   }
#else
   // assign data buffers with malloc using kernel memory
   for(n=0; n<(adc->nbufs); n++) {
      adc->buffer[n] = (char *)malloc(nbytes); 
      if(adc->buffer[n] == NULL) {
         fprintf(err_fp, "Error allocating adc buffers\n"); 
         return(-1);
      }
   }
#endif

   if(debug_level > 1) {
     for(n=0; n<(adc->nbufs); n++) {
         fprintf(dbg_fp,"ad7766_alloc_dma_buffers: adc buffer[%d] = %p\n", n, adc->buffer[n]);
     }
   }

   return(adc->nsamps);
}

//
// Setup ADC clocks.  
// Timer0 is used for thet adc mclk. The serial bus uses timer1 as rsclk 
// Timer1 (rsclk) is synced to timer0 (mclk)
//
int ad7766_start_clock(adc_t *adc)
{
   unsigned long mclk, R, rsclk, sclk;
   unsigned long period, width;
   int fs;
   //float t4, t5;

   fs = adc->fs;
   R = 8;
   mclk = R*(unsigned long)fs;   // ad7766 fs = mclk/8

   // use a timer as the mclk signal
   mclk_fd = open("/dev/timer0", O_RDONLY, 0);
   if (mclk_fd < 0) {
      fprintf(err_fp, "ad7766_clock: Failed to open timer0\n");
      return(-1);
   } 

   // set mclk signal
   ioctl(mclk_fd, BFIN_TIMER_OUTPUT_MODE);
   if (ioctl(mclk_fd, BFIN_TIMER_SET_FREQ, &mclk) < 0) {
      fprintf(err_fp, "ad7766_clock: failed to set MCLK\n");
      return(-1);
   } 

   // get sclk 
   if (ioctl(mclk_fd, BFIN_TIMER_GET_SCLK, &sclk)  < 0) {
      fprintf(err_fp, "ad7766_clock: failed to get SCLK\n");
      return(-1);
   } 

   // actual sampling freq
   fs = mclk/R;  
   adc->fs = fs; 

   // rsclk is 5 times the mclk - found by testing
   rsclk = 5*mclk;  // rcv serial clk
   period = sclk/rsclk;
   width = period/2;

   // use a timer as the rsclk signal
   rsclk_fd = open("/dev/timer1", O_RDONLY, 0);
   if (rsclk_fd < 0) {
      fprintf(err_fp, "ad7766_clock: Failed to open timer1\n");
      return(-1);
   } 

   // set rsclk signal
   ioctl(rsclk_fd, BFIN_TIMER_OUTPUT_MODE);
   ioctl(rsclk_fd, BFIN_TIMER_SET_PERIOD, period);
   if (ioctl(rsclk_fd, BFIN_TIMER_SET_WIDTH, width) < 0) {
      fprintf(err_fp, "ad7766_clock: failed to set RSCLK\n");
      return(-1);
   } 

   if(debug_level > 0) {
      fprintf(dbg_fp, "ad7766_clock: fs=%d, mclk=%ld, rsclk=%ld, sclk=%ld\n", 
         fs, mclk, rsclk, sclk);
      if(debug_level > 1) {
         fprintf(dbg_fp, "ad7766_clock: rsclk period=%ld, width=%ld\n", period,width);
      }
   }

   // start synchonized mclk and rsclk 
   ioctl(mclk_fd, BFIN_TIMER_PULSE_HI);
   ioctl(rsclk_fd, BFIN_TIMER_PULSE_HI);
   if (ioctl(mclk_fd, BFIN_TIMER_START_SYNC, 1) < 0) {
      fprintf(err_fp, "ad7766_clock: failed to START\n");
      return(-1);
   } 

   if (ioctl(mclk_fd, BFIN_TIMER_START) < 0) {
      fprintf(err_fp, "ad7766_clock: failed to START\n");
      return(-1);
   } 

   return(1);
}


int ad7766_start_dma(adc_t *adc)
{
   struct sport_dma_config dma_config;
   int n, bufsize, nbufs, xcount, ycount, word_len;
   
   bufsize = adc->nsamps;
   nbufs = adc->nbufs;
   word_len = 32;  // always 3 2bits

   // find dma xcount and ycount
   xcount = 0xFFFF; 
   ycount = bufsize / xcount;
   while((xcount > 1) & ((xcount*ycount) != bufsize)) {
      xcount = xcount - 1;
      ycount = bufsize / xcount;
   }

   if(debug_level > 0) {
     fprintf(dbg_fp,"ad7766_start_dma: xcount = %d, ycount=%d\n", xcount, ycount);
   }

   // Configure dma buffers
   dma_config.mode = DMA_RX_QUEUE; 
   dma_config.word_len = (char)word_len;
   dma_config.row_size = (short)xcount;
   dma_config.num_buffers = nbufs;
   dma_config.buffer_size = bufsize;
   for (n=0; n<nbufs; n++) {
      if(adc->buffer[n] == NULL) {
         fprintf(dbg_fp,"ad7766_start_dma: invalid buffer pointer.\n");
		 return(-1);
	  }
      dma_config.buffer[n] = (unsigned long)(adc->buffer[n]);
   }

   /* Configure sport dma controller by ioctl */
   if (ioctl(adc->fd, SPORT_DMA_CONFIG, &dma_config) < 0) {
      fprintf(err_fp, "ad7766_start_dma: failed to config sport dma\n");
      return(-1);
   }

   //ioctl(adc->fd, SPORT_DMA_START);
   //ioctl(adc->fd, SPORT_DMA_STOP);
   //ioctl(adc->fd, SPORT_DMA_RX_CONFIG, &dma_config);

   /* Configure sport dma queue */
   if (ioctl(adc->fd, SPORT_DMA_START) < 0) {
      fprintf(err_fp, "ad7766_start_dma: failed to start sport dma\n");
      return(-1);
   }

   return(1);

}

int ad7766_stop_dma(adc_t *adc)
{
   /* Configure sport dma queue */
   if (ioctl(adc->fd, SPORT_DMA_STOP) < 0) {
      fprintf(err_fp, "ad7766_stop_dma: failed to stop sport dma\n");
      return(-1);
   }

   // stop mclk
   if (ioctl(mclk_fd, BFIN_TIMER_STOP) < 0) {
      fprintf(err_fp, "ad7766_stop_dma: failed to STOP\n");
      return(-1);
   } 

   return(1);
}

// 
// Check if a dma buffer is ready.
// If a buffer is ready to read, return the number of bytes.
// The buffer is available in the at adc->buffer.addr
//
int ad7766_read_dma(adc_t *adc, char **buffer)
{
   int nsamps;
   struct timeval now;

   /* Configure sport dma queue */
   if (ioctl(adc->fd, SPORT_DMA_READ, &(adc->dma_buffer)) < 0) {
      fprintf(err_fp, "ad7766_read_dma: failed to read sport dma queue\n");
      return(-1);
   }
   
   *buffer = NULL;

   if(adc->dma_buffer.done == 0) {
      return(0);
   }

   // set parameters for return
   //sec = (double)(tv.tv_sec) + (double)(tv.tv_usec)*0.000001;
   adc->time.tv_sec = adc->dma_buffer.tv.tv_sec;
   adc->time.tv_nsec = adc->dma_buffer.tv.tv_nsec;
   
   // if there's no time stamp on the dma buffer use current time
   if(adc->time.tv_sec == 0) {
	   gettimeofday(&now, NULL);
	   adc->time.tv_sec = now.tv_sec;
	   adc->time.tv_nsec = now.tv_usec*1000;
   }
   
   nsamps = adc->dma_buffer.size;
   
   // set pointer to finished buffer
   *buffer = (char *)(adc->dma_buffer.addr);
   
   if(debug_level > 1) {
     fprintf(dbg_fp,"ad7766_read_queue: sec=%ld, nsec=%ld\n", 
       adc->time.tv_sec, adc->time.tv_nsec);
   }

   return(nsamps);
}

/*
 * Read ad7766 data into a 32 bit word buffer.
 * The word is read as 32 bits with the first 24 bits of data
 * and 8 bits of padding. 
 * The buffer needs to be allocated as a 32 bit word.
*/
int ad7766_read(adc_t *adc, char *buf, int nsamps, int nsize)
{
   int nrd;
   int nbytes;
   struct timeval tv;
   //double sec;

   // reset adc
   //gpio_write(gpio_adc_reset, '0');
   //usleep(100);
   //gpio_write(gpio_adc_reset, '1');
   //usleep(1000);

   nbytes = nsize * nsamps; 

   // time stamp as close to read as possible
   //gettimeofday(&tv, &tz); 
   gettimeofday(&tv, NULL); 

   // read data from sport
   nrd = read(adc->fd, buf, nbytes);
   if (nrd < nbytes) {
      fprintf(err_fp, "ad7766_read: error: %d\n", nrd);
      return(-1);
   } 

//   ioctl (mclk_fd, BFIN_TIMER_STOP);
//   ioctl (rsclk_fd, BFIN_TIMER_STOP);

   // set parameters for return
   //sec = (double)(tv.tv_sec) + (double)(tv.tv_usec)*0.000001;
   adc->time.tv_sec = tv.tv_sec;
   adc->time.tv_nsec = tv.tv_usec/1000;
   adc->nsamps = nsamps;
   adc->nsize = nsize;
   adc->nchans = 1;
   nrd = nrd/nsize;

   if(debug_level > 1) {
     fprintf(dbg_fp,"ad7766_read: nsamps=%d, sec=%ld.%03ld\n", 
       nrd, tv.tv_sec, tv.tv_usec/1000);
   }

   return(nrd);
}

/*
 Reformat the 24 bit word read from the ad7766 into a 16 bit word by
 shifting the 24 bit word right (>>) by a specified number of bits (shift).
 The value of shift can be between 0 and 8.
 If shift = 0, then the 16 least significant bits of the 24 bit adc word are save.
 If shift = 8, then the 16 most significant bits of the 24 bit adc word are save. 
 The adc buffer has 32 bit words with valid data in the last 24 bits.
 Bit 24 is a sign bit and the word is a two's compliment integer.
*/
int ad7766_scale_16(char *buf, int nsamps, int shift)
{
   int n;
   unsigned short *buf16;
   unsigned long *buf32;
   if(shift>8) shift=8;
   buf32 = (unsigned long *)buf;
   buf16 = (unsigned short *)buf;
   for(n=0; n<nsamps; n++) {
      buf16[n] = (unsigned short)(buf32[n] >> shift);
   }
   return(0);
}

/*
 * Reformat 32 bit word into 24 but word.
 */
int ad7766_scale_24(char *buf, int nsamps)
{
   unsigned long n, i, j;
   unsigned char *buf32;
   unsigned char *buf24;
   buf32 = (unsigned char *)buf;
   buf24 = (unsigned char *)buf;
   for(n=0; n<nsamps; n++) {
	   i = n*4;
	   j = n*3;
       buf24[j] = buf32[i+1]; 
       buf24[j+1] = buf32[i+2]; 
       buf24[j+2] = buf32[i+3]; 
   }
   return(0);
}

/*
 Reformat the 24 bit word read from the adc into a 32 bit word by 
 shifting left (<<) 8 bits to put the sign bit in the MSB.
 The adc buffer has 32 bit words with valid data in the last 24 bits.
 Bit 24 is a sign bit and the word is a two's compliment integer.
*/
int ad7766_scale_32(char *buf, int nsamps)
{
   int n;
   unsigned long *buf32;
   buf32 = (unsigned long *)buf;
   for(n=0; n<nsamps; n++) {
      buf32[n] = buf32[n] << 8; 
   }
   return(0);
}

int ad7766_scale(adc_t *adc, char *buf)
{
   int nsamps, shift;

   nsamps = adc->nsamps;
   shift = adc->shift;
   if(adc->nsize == 2) {
      ad7766_scale_16(buf, nsamps, shift);
   }
   else if(adc->nsize == 3) {
      ad7766_scale_24(buf, nsamps);
   }
   else if(adc->nsize == 4) {
      ad7766_scale_32(buf, nsamps);
   }
   return(0);
}

void ad7766_close(adc_t *adc)
{
   // turn ADC VREF OFF
   gpio_write(GPIO_VREF, 0);
   gpio_unexport(GPIO_VREF);

   // stop mclk signal
   ioctl (mclk_fd, BFIN_TIMER_STOP);
   close(mclk_fd);
   ioctl (rsclk_fd, BFIN_TIMER_STOP);
   close(rsclk_fd);

#ifndef ADC_USE_PMEM
   int n;
   for(n=0; n<(adc->nbufs); n++) {
      free(adc->buffer[n]);
   }
#endif

   close(adc->fd);
   //free(adc);
}

