/*
 *  com.c: serial com interface
 *
 *  Combination of tty command console and message handling functions
 *  For docs on how to set serial port see 
 *    http://www.easysw.com/~mike/serial/serial.html
 *
 * Commands strings start with a $ and end with a *.
 *
 * ------
 * Embedded Ocean 2013
*/

#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>

#include "wispr.h"
#include "com.h"
#include "rtc.h"
#include "misc.h"

//#define DEBUG

static int tty_fd;
static int tty_buffer_len;
static int tty_buffer_full;
static char tty_buffer[COM_BUFSIZE];
static struct termios tio;
static struct termios otio;	// will be used to save old port settings
static struct sigaction saio;	// set the serial interrupt handler

// interupt signal handler
void com_signal_handler (int status);

extern int debug_level;
extern FILE *dbg_fp;
extern FILE *err_fp;

//---------------------------------------------------------
/* Open com port
 * Serial port is used in cannonical mode
*/

int com_open (char *device, speed_t baudrate)
{
  int fd;

  // open tty console
  fd = open (device, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) {
    fprintf (err_fp, "com_open: unable to open serial port (%s)\n",
	     device);
    return (0);
  }

  tcgetattr (fd, &tio);		// get current port settings
  tcgetattr (fd, &otio);	// save current port settings
  cfsetispeed (&tio, baudrate);
  tio.c_cflag &= ~PARENB;	// no parity
  tio.c_cflag &= ~CSTOPB;	// one stop bit
  tio.c_cflag &= ~CSIZE;	/* Mask the character size bits */
  tio.c_cflag |= CS8;		/* Select 8 data bits */
  tio.c_cflag &= ~CRTSCTS;	// disable hdwr flow cntrl
  tio.c_lflag |= (ICANON);	// canonical input
  tio.c_lflag &= ~(ECHO | ECHOE);	// canonical input
  //tio.c_lflag |= (ICANON | ECHO | ECHOE); // canonical input
  tio.c_lflag |= (ISIG);	// enable input signal
  //tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // raw input
  //tio.c_iflag |= (INPCK | ISTRIP); // parity checks
  //tio.c_iflag |= (ICRNL); // map CR to NL
  tio.c_iflag &= ~(IXON | IXOFF | IXANY);	// disable soft flow cntrl
  tio.c_oflag &= ~OPOST;	// raw output
  //tio.c_cc[VMIN] = 0;   // use-vtim, raw input onlye
  //tio.c_cc[VTIME] = 1;  // time to wait until exiting read

  /* now clean the modem line and activate the settings for the port */
  tcflush (fd, TCIFLUSH);	// flush old data
  tcsetattr (fd, TCSANOW, &tio);	// apply new settings
  //tcflush(fd, TCIOFLUSH);

  /* allow the process to receive SIGIO */
  fcntl (fd, F_SETOWN, getpid ());
  /* Make the file descriptor asynchronous */
  fcntl (fd, F_SETFL, FASYNC);

  /* install the signal handler before making the device asynchronous */
  saio.sa_handler = com_signal_handler;
  sigemptyset (&saio.sa_mask);	// clear existing settings
  saio.sa_flags = 0;
  saio.sa_restorer = NULL;
  sigaction (SIGIO, &saio, NULL);

  // save local vriables
  tty_fd = fd;
  //strncpy (tty_device, device, 32);
  bzero (tty_buffer, COM_BUFSIZE);	// Initialize to all zeros
  tty_buffer_full = 0;

  return (fd);
}

//---------------------------------------------------------
/* Interupt function is called whenever there is new data 
 * to be read off the serial port.
 * The interupt routine reads max buffer size from serial port device 
 * (cannonical mode) and puts the message into a buffer (tty_buffer).
 * It has to execute quickly, so the data processing is not done here.
*/
void com_signal_handler (int status)
{
  char buf[COM_MSGSIZE];
  int n, m;

  // read port
  n = read (tty_fd, buf, COM_MSGSIZE);
  //fprintf(dbg_fp,"com sig: %d\n", n);

  if (n > 0) {
    if (buf[0] == '$') {	// it's a valid message
      //buf[n-1] = '*'; //  add terminating char
      buf[n - 1] = 0x00;	// replace <CR>
      m = strlen (tty_buffer);
      //fprintf (dbg_fp, "com sig: %d, %d, %s\n", m, n, buf);
      if ((m + n) < COM_BUFSIZE - 1) {
	// append new data to the buffer of unprocessed data
	strcat (tty_buffer, buf);
	tty_buffer_len += n;
	tty_buffer_full = 0;
      }
      else {
	// buffer full
	tty_buffer_full = 1;
	fprintf (err_fp, "com_signal_handler: buffer full\n");
      }
      //fprintf(dbg_fp,"com sig: %d, buffer %s\n", 
      //   strlen(tty_buffer), tty_buffer);
    }
  }
}

//---------------------------------------------------------
int com_close ()
{
  // restore the old port settings before quitting
  //tcflush(tty_fd, TCIFLUSH);      // flush data
  tcsetattr (tty_fd, TCSANOW, &(otio));
  close (tty_fd);
  return 0;
}

//---------------------------------------------------------
/*
* Copy next message in the tty_buffer into msg buffer
* Copy chars between the next '$' and '*' chars.
* Shifts the rest of the buffer forward.
* Removes any junk in the buffer (text without a $ prefix).
*/
int com_read_msg (char *msg, int msglen)
{
  int n, m;
  char tmp[COM_BUFSIZE], *start, *tail;

  // return if no msg in buffer
  n = strlen (tty_buffer);
  if (n <= 0) return (0);

  // make a copy
  strcpy (tmp, tty_buffer);

  // find start, end, and size of first msg in buffer
  start = strchr (tmp, '$');	// find start of msg
  tail = strchr (tmp, '*');	// find end of msg
  m = (int) (tail - start - 1);	// size of message, less '$' and '*' chars
  if (m <= 0) return (0);

  // copy first message into msg buffer, skip '$' char
  strncpy (msg, start + 1, m);

  // terminate msg
  msg[m] = 0x00;

  // copy the rest back onto buffer
  strcpy (tty_buffer, tail + 1);

  //fprintf (dbg_fp, "com_read_msg: %s\n", msg);
  //fprintf(dbg_fp, "start=%d, tail=%d, m=%d\n", start, tail, m);

  return (m);
}

//---------------------------------------------------------
int com_write_msg (char *msg, int n)
{
  int len, m;
  char obuf[COM_MSGSIZE];
  obuf[0] = '$';
  strncpy (obuf + 1, msg, n);
  obuf[n + 1] = '*';
  //obuf[n+2] = 0x0d;   // <CR> after msg 
  obuf[n + 2] = 0x0a;	// <LF> newline 
  obuf[n + 3] = 0x00;	// terminate the string
  len = strlen (obuf);
  //len = n+2; 
  m = write (tty_fd, obuf, len);
  if(debug_level > 0) fprintf (dbg_fp, "com_write_msg: %d, %s", m, obuf);
  if (m < 0) {
    fprintf (err_fp, "com_write_msg: error%d, %s", m, obuf);
    return 0;
  }
  //tcflush(tty_fd, TCIFLUSH);      // flush data
  return (m);
}

//---------------------------------------------------------
int com_write_raw (char *msg, int len)
{
  int m;
  char obuf[COM_MSGSIZE];
  strncpy (obuf, msg, len);
  m = write (tty_fd, obuf, len);
  //fprintf(dbg_fp,"tty_write_msg: %d, %s\n", m, obuf);
  if (m < 0) {
    fputs ("com_write_raw: failed!\n", stderr);
    return 0;
  }
  //tcflush(tty_fd, TCIFLUSH);      // flush data
  return (m);
}

//---------------------------------------------------------
int com_parse_msg (char *buf, int len, com_t * com)
{
  int mode;
  char msg[6], attr[128], tmstr[32];
  float lat, lon;
  time_t sec;
  float press, depth, rho, g;
  char rtc_device[32];

  strcpy(rtc_device, "/dev/rtc0");

  sec = 0;
  press = 0.0;			// pressure N/m^2
  rho = 1025.0;			// seawater density kg/m^3
  g = 9.81;			// gravity m/s^2 

  // remember that there is no '$' at the front of buf
  msg[0] = buf[0];
  msg[1] = buf[1];
  msg[2] = buf[2];
  msg[3] = 0;

  buf[len + 1] = 0;
  strcpy (attr, &buf[4]);

  if (debug_level) {
    fprintf (dbg_fp, "com_parse_msg: %s,%s\n", msg, attr);
  }

  // Piston 
  if (strncmp (msg, "PMA", 3) == 0) {
    sscanf (attr, "%ld,%d", &sec, &mode);
    if (mode == 1) {
      com->mode |= COM_PISTON;
      com->state = COM_PAUSE;
    } else if (mode == 0) {
      com->mode &= ~COM_PISTON;
      com->state = COM_RUN;
    }
  }
  // Airbag
  if (strncmp (msg, "ABS", 3) == 0) {
    sscanf (attr, "%ld,%d", &sec, &mode);
    if (mode == 1) {
      com->mode |= COM_AIRBAG;
      com->state = COM_PAUSE;
    } else if (mode == 0) {
      com->mode &= ~COM_AIRBAG;
      com->state = COM_RUN;
    }
  }
  // Descend
  if (strncmp (msg, "DSC", 3) == 0) {
    sscanf (attr, "%ld,%f", &sec, &press);
    com->state = COM_RUN;
    com->mode &= ~COM_ASCENDING;
    com->mode |= COM_DESCENDING;
  }
  // Ascend
  if (strncmp (msg, "ASC", 3) == 0) {
    sscanf (attr, "%ld,%f", &sec, &press);
    com->state = COM_RUN;
    com->mode |= COM_ASCENDING;
    com->mode &= ~COM_DESCENDING;
  }
  // Park
  if (strncmp (msg, "PRK", 3) == 0) {
    sscanf (attr, "%ld,%f", &sec, &press);
    com->state = COM_RUN;
    com->mode |= COM_PARKED;
    com->mode &= ~COM_PROFILING;
  }
  // Profiling mode
  if (strncmp (msg, "PFL", 3) == 0) {
    sscanf (attr, "%ld,%f", &sec, &press);
    com->state = COM_RUN;
    com->mode |= COM_PROFILING;
    com->mode &= ~COM_PARKED;
  }
  // Surface mode
  if (strncmp (msg, "SDT", 3) == 0) {
    sscanf (attr, "%ld,%f", &sec, &press);
    com->state = COM_PAUSE;
    com->mode |= COM_SURFACE_DETECT;
  }

  // Exit
  if (strncmp (msg, "EXI", 3) == 0) {
    com->state = COM_EXIT;
  }
  // Run
  if (strncmp (msg, "RUN", 3) == 0) {
    com->state = COM_RUN;
  }
  // Pause
  if (strncmp (msg, "PAU", 3) == 0) {
    com->state = COM_PAUSE;
  }
  // Reset
  if (strncmp (msg, "RES", 3) == 0) {
    com->state = COM_RESET;
  }
  // Sleep
  if (strncmp (msg, "SLP", 3) == 0) {
    com->state = COM_SLEEP;
  }

  // Actions
  // GPS message
  if (strncmp (msg, "GPS", 3) == 0) {
    sscanf (attr, "%ld,%f,%f", &sec, &lon, &lat);
    com->gps->time = sec;
    com->gps->lon = lon;
    com->gps->lat = lat;
    rtc_set_time(rtc_device, sec); // set RTC time
    set_system_time(sec, 0);	// set linux system time here
    com->mode |= COM_GPS;	// new gps update 
    if(debug_level) {
      ctime_r(&sec, tmstr);
      fprintf (dbg_fp, "com_parse_msg: gps, %s", tmstr);
      fprintf (dbg_fp, "com_parse_msg: gps, lat=%f, lon=%f\n", lat, lon);
    }
  }

  // set time, if new time avaiable
  if (sec > 0) {
    com->time = sec;
    com->mode |= COM_TIME;	// new time value
  }

  // set depth if new pressure avaiable
  if (press > 0.0) {
    depth = press / (rho * g);
    com->gps->depth = depth;
    com->mode |= COM_DEPTH;	// new depth value
  }

  // save message to log file
  if(com->file != 0) {
    FILE *fp = fopen(com->file, "a+");
    if(fp == NULL) 
      fprintf(err_fp, "com_parse_msg: error opening log file\n");
    else {
      fprintf(fp, "$%s*\n", buf);
      fclose(fp);
    }
  }

  if (debug_level) {
    fprintf (dbg_fp, "com_parse_msg: mode=0x%x, state=%d\n",
      com->mode, com->state);
  }

  return (com->state);
}

// ------------------------------------------------------------------------

pthread_mutex_t com_mutex = PTHREAD_MUTEX_INITIALIZER;

void *com_handle_msg (void *arg)
{
  com_t *com = (com_t *) arg;

  int nrd;
  char device[32];
  int fd;
  char msg[COM_MSGSIZE];
  int state, mode;

  strcpy (device, "/dev/ttyBF1");
  
  fd = com_open (device, B9600);
  if (fd < 0) {
    //perror ("com_open");
    fprintf(err_fp, "com_handle_msg: error opening tty device\n");
    return (0);
    //exit(EXIT_FAILURE);
  }

  state = COM_RUN;
  mode = 0;
  pthread_mutex_lock (&com_mutex);
  com->state = state;
  com->mode = mode;
  pthread_mutex_unlock (&com_mutex);

  while (state > 0) {

    // check for serial port actions
    nrd = com_read_msg (msg, COM_MSGSIZE);

    // parse msg , if available
    pthread_mutex_lock (&com_mutex);
    if (nrd > 0) {
      state = com_parse_msg (msg, nrd, com);
    }
    state = com->state;
    mode = com->mode;
    pthread_mutex_unlock (&com_mutex);

    // actions
    if (state == COM_EXIT) break;
    else sleep (1);

  }

  com_close ();

  return (NULL);
}


