/*
 * file.c - data file I/O utilities
 *
 * Data files are straight binary with no headers.
 * For each raw data file there's a separate ascii header file with the same base name.
 * The header file contains the description for the raw data. 
 * File names contain a user defined prefix and the date:
 * Raw data file names have a .dat extension (prefix_YYMMDD_HHMMSS.dat)
 * Header file names have a .txt extension (prefix_YYMMDD_HHMMSS.txt)
 * Wave data file format can also be used.
 * ------
 * Embedded Ocean Systems 2013
 */

#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/statfs.h>

#include <cycle_count.h>
#include <time.h>

#include "file.h"
#include "wave.h"
#include "misc.h"

extern int verbose;
extern int debug_level;
extern FILE *dbg_fp;
extern FILE *err_fp;

/*
 * write data to file
 * makes a new file if current file is full
 */
int file_write(file_t *file, data_t *data)
{
   char path[32], name[32], prefix[8], ext[8], str[128];
   struct tm time;
   uint32 sec, usec;
   int nwrt=0;
   int nblks;
   //struct statvfs stat;
   struct statfs stat;

   strcpy(path, file->path);
   strcpy(name, file->name);
   strcpy(prefix, file->prefix);
   sec = (uint32)(data->time.tv_sec);
   usec = (uint32)(data->time.tv_usec);

   if(file->bfree <= file->bfree_min) {
      if(verbose) print_log(file, "file system full");
      if(debug_level > 0) {
         fprintf(dbg_fp, "file_write: only %ld blocks free, no data written.\n", 
            file->bfree);
      }
      return(0); 
   }

   if(file->format == WAVE_FILE_FORMAT) {  // wave format
      strcpy(ext,"wav"); 
   } else { // default format
      strcpy(ext,"dat"); 
   }

   // make a new data file, 
   // if maxblks is exceeded (file is full), or if no blocks have been written yet (file is new),
   if((file->nbufs == 0) || (file->nbufs >= file->max_nbufs)) {
	    
      // make new file name from the prefix and date
      file->open_sec = sec;
      gmtime_r((time_t *)&sec, &time);
      strftime(str, sizeof(str), "%g%m%d_%H%M%S", &time); 
      sprintf(name, "%s%s", prefix, str);
      
	  strcpy(file->name, name);

      file->nbytes = 0;
      file->nbufs = 0;

      // check disk space
      // - statfs could be slow and cause buffer over runs when a new file is created.
      if((statfs(path, &stat)) < 0 ) {
         fprintf(err_fp, "file_write: Failed to stat %s:\n", path);
      } else {
         file->bfree = (unsigned long)(stat.f_bavail);
         file->blocks = (unsigned long)(stat.f_blocks);
      }

      // write data file header for new file
      nwrt = file_write_hdr(file, data); 

      snprintf(str, sizeof(str), "New file %s, %0.2f%% fs used",
         file->name, 100.0*(float)(file->blocks - file->bfree)/(float)(file->blocks));
      print_log(file, str);

   }

   // write the data buffer in the specified format
   if(file->format == WAVE_FILE_FORMAT) {  // wave format
      nwrt = file_append_wave(file, data); 
   } else { // default format
      nwrt = file_append(file, data);
   }

   // increment number of data buffers written to file
   file->nbytes += nwrt;
   file->nbufs++;

   // decrement number of free blocks in fs 
   nblks = nwrt / file->bsize;
   file->bfree = file->bfree - nblks;

   //curr_time = (double)(data->time.tv_sec) +
   //            (double)(data->time.tv_usec)/1000000.0;

   if(verbose > 1) {
      sprintf(str, "buffer %d written", file->nbufs);
      if(data->status & DATA_CLIPPED) strcat(str,", clipped");
      if(data->status & DATA_OVERRUN) strcat(str,", overflow");
      print_log(file, str);
   }

   //last_time = curr_time;

   if(debug_level > 1) {
      fprintf(dbg_fp, "file_write: %d bytes written, %ld blks free\n",
         file->nbytes, (file->bfree));
      // check disk space
      //if(debug_level > 2) {
      //   if((statfs(path, &stat)) < 0 ) {
      //      fprintf(err_fp, "file_write: Failed to stat %s:\n", path);
      //   }
      //   fprintf(dbg_fp, "file_write: f_bavail = %ld\n", (stat.f_bavail));
      //}
   }

   return(nwrt);
}

/*
 * write header information to a text file
 */
int file_write_hdr(file_t *file, data_t *data)
{
   char path[32], name[32], filename[66];
   uint32 nsamps, fs, sec, usec, nblks;
   uint16 nchans, nsize, gain, scale;
   int nwrt=0;
   FILE *fp;

   // define header parameters
   // the hdr.nblks is the number of data buffers in the file
   // the hdr.nsamps is the number of samples in each buffer
   strcpy(path, file->path);
   strcpy(name, file->name);
   nblks = (uint32)(file->max_nbufs);
   sec = (uint32)(data->time.tv_sec);
   usec = (uint32)(data->time.tv_usec);
   nsamps = (uint32)(data->nsamps); // total buffer size
   nchans = (uint16)(data->nchans);
   fs = (uint32)(data->fs);
   nsize = (uint16)(data->nsize);
   gain = (uint16)(data->gain);
   scale = (uint16)(data->scale);

   // open current data file for appending
   sprintf(filename, "%s/%s.txt", path, name);
   //fd = open(filename, O_WRONLY|O_EXCL|O_CREAT, 0);
   fp = fopen(filename, "w+");
   if(fp <= 0) {
      fprintf(err_fp, "file_write_hdr: Can't open output file, no data written\n");
      return(-1);
   }

   // write the header file text
   fprintf(fp, "%% WISPR DATA FILE HEADER\n");
   fprintf(fp, "file name: %s\n", name);
   fprintf(fp, "start time sec: %ld\n", sec);
   fprintf(fp, "start time usec: %ld\n", usec);
   fprintf(fp, "samples per second: %ld\n", fs);
   fprintf(fp, "blocks per file: %ld\n", nblks);
   fprintf(fp, "samples per block: %ld\n", nsamps);
   fprintf(fp, "bytes per sample: %d\n", nsize);
   fprintf(fp, "channels: %d\n", nchans);
   fprintf(fp, "gain: %d\n", gain);
   fprintf(fp, "scale: %d\n", scale);

   // write 64 byte data block header to file
   //fprintf(dbg_fp, "writing hdr\n");
   //int n;
   //char hdr[8];
   //for(n=0; n<64; n++) hdr[n]=0;
   //hdr[0] = 'W'; hdr[1] = 'I'; hdr[2] = 'S'; hdr[3] = 'P'; hdr[4] = 'R';
   //memcpy(&hdr[4],  &sec,    4);
   //memcpy(&hdr[8],  &usec,   4);
   //memcpy(&hdr[12], &nblks,  4);
   //memcpy(&hdr[16], &nsamps, 4);
   //memcpy(&hdr[20], &nchans, 2);
   //memcpy(&hdr[22], &nsize,  2);
   //memcpy(&hdr[24], &fs,     4);
   //memcpy(&hdr[28], &gain,   2);
   //memcpy(&hdr[30], &scale,  2);
   //nwrt = fwrite(hdr, 1, 64, fp);
   ////nwrt = write(fd, hdr, 32);

   fclose(fp);

   return(nwrt);
}

/*
 * read file header
 */
//int file_read_hdr(file_t *file, data_t *data)
int file_read_hdr(FILE *fp, header_t *hdr)
{
   uint32 nsamps, fs, sec, usec, nblks;
   uint16 nchans, nsize, gain, scale;
   int nrd=0;
   char buf[65];

   // read 64 byte data block header
   nrd = fread(buf, 1, 64, fp);

   memcpy(&sec, &buf[4],  4);
   memcpy(&usec, &buf[8],  4);
   memcpy(&nblks, &buf[12], 4);
   memcpy(&nsamps, &buf[16], 4);
   memcpy(&nchans, &buf[20], 2);
   memcpy(&nsize, &buf[22], 2);
   memcpy(&fs, &buf[24], 4);
   memcpy(&gain, &buf[28], 2);
   memcpy(&scale, &buf[30], 2);

   hdr->time.tv_sec = sec;;
   hdr->time.tv_usec = usec;
   hdr->nblks = nblks;
   hdr->nsamps = nsamps; // total buffer size
   hdr->nchans = nchans;
   hdr->fs = fs;
   hdr->nsize = nsize;
   hdr->gain = gain;
   hdr->scale = scale;

   return(nrd);
}

/*
 * write data to raw file
 */
int file_append(file_t *file, data_t *data)
{
   char path[32], name[32], filename[66];
   size_t bufsize, nsamps, nchans, nsize;
   size_t nbytes, bsize;
   void *buf;
   int nwrt=0;
   FILE *fp;

   strcpy(path, file->path);
   strcpy(name, file->name);

   nsamps = (size_t)(data->nsamps);
   nchans = (size_t)(data->nchans);
   nsize = (size_t)(data->nsize);
   bufsize = nsize*nsamps*nchans;
   buf = (void *)(data->buffer);

   // open current data file for appending
   sprintf(filename, "%s/%s.dat", path, name);
   fp = fopen(filename, "a+");
   if(fp <= 0) {
      fprintf(err_fp, "file_append: Can't open output file, no data written\n");
      return(-1);
   }

   // write data buffer
   //nwrt = fwrite(buf, 1, bufsize, fp);
   bsize = 4096;
   nbytes = nsize*nsamps;
   nwrt = bwrite(buf, bsize, nbytes, fp);
   if(nwrt < nbytes) {
      fprintf(stderr, "file_append: error writing data to %s\n", name);
   }

   // force output to file
   fflush(fp);
   fclose(fp);

   return(nwrt);
}

/*
 * append data to wave file
 */
int file_append_wave(file_t *file, data_t *data)
{
   char path[32], name[32], filename[66];
   uint32 nsamps, fs, sec, usec;
   uint16 nchans, nsize, gain, scale;
   void *buf;
   int nwrt=0;
   struct WaveFile wave;

   strcpy(path, file->path);
   strcpy(name, file->name);

   sec = (uint32)(data->time.tv_sec);
   usec = (uint32)(data->time.tv_usec);
   nsamps = (uint32)(data->nsamps);
   nchans = (uint16)(data->nchans);
   fs = (uint32)(data->fs);
   nsize = (uint16)(data->nsize);
   gain = (uint16)(data->gain);
   scale = (uint16)(data->scale);
   buf = (void *)(data->buffer);

   strcpy(wave.ChunkID,"RIFF");
   strcpy(wave.Format,"WAVE");
   strcpy(wave.Subchunk1ID,"fmt"); 
   strcpy(wave.Subchunk2ID,"data");
   wave.Subchunk1Size = 16;
   wave.Subchunk2Size = nsamps * nsize;
   wave.ChunkSize = 4 + (8 + wave.Subchunk1Size) + (8 + wave.Subchunk2Size);
   wave.AudioFormat = 1;
   wave.NumChannels = nchans;
   wave.SampleRate = fs;
   wave.ByteRate = fs * nchans * nsize;
   wave.BlockAlign = nchans * nsize;
   wave.BitsPerSample = 8*nsize;

   // appending
   sprintf(filename, "%s/%s", path, name);
   nwrt = wave_append(filename, &wave, buf, nsamps, nsize);

   return(nwrt);
}



