/*
 * WISPR data logger
 * ------
 * Embedded Ocean, Sept 2013
 */

#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <linux/rtc.h>
#include <sys/wait.h>
#include <assert.h>
#include <pthread.h>
#include <fract.h>
#include <filter.h>
#include <stats.h>
#include <math.h>
#include <cycle_count.h>
#include <time.h>
#include <sys/mman.h>
#include <linux/watchdog.h>
#include <sys/statfs.h>

#include "wispr.h"
#include "ad7766.h"
#include "com.h"
#include "gpio.h"
#include "rtc.h"
#include "wdt.h"
#include "file.h"
#include "wave.h"
#include "led.h"
#include "misc.h"

#define GPIO_SD GPIO_PG0
#define GPIO_G1 GPIO_PG1
#define GPIO_G0 GPIO_PG2

char *progname;
int verbose = 0;
int debug_level = 0;
FILE *dbg_fp;
FILE *err_fp;

int set_adc_gain(int gain); 


void print_banner(file_t *file, data_t *dat)
{
   float duration, total_duration;
   char str[128];

   print_log(file, "WISPR STARTED");

   sprintf(str, "File system: %0.2f MB of %0.2f MB (%0.2f%%) available",
      ((float)(file->bfree)*(float)(file->bsize)/1000000.0),
      ((float)(file->blocks)*(float)(file->bsize)/1000000.0),
      100.0*(float)(file->bfree)/(float)(file->blocks));
   print_log(file, str);

   //fprintf(dbg_fp, " %d blocks free.\n", (int)(stat.f_bfree));
   duration = (float)dat->nsamps/(float)dat->fs; // record duration in seconds
   sprintf(str, "nsamps=%d (%f sec), fs=%d, gain=%d, shift=%d",
      dat->nsamps, duration, dat->fs, dat->gain, dat->scale);
   print_log(file, str);

   total_duration = ((float)(file->bfree)*(float)(file->bsize)/2.0)/(float)dat->fs;
   sprintf(str, "Recording time %0.2f hours", total_duration/(3600.0));
   print_log(file, str);

}

int main(int argc, char **argv)
{
   char *progname, str[128]; 
   int rtn, nrd, nwrt, opt, ndtx;
   char adc_device[32];
   char tty_device[32];
   int state, prev_state, mode;
   unsigned long blksize;
   int count, max_count;
   char *active_buffer;
   com_t com;
   char com_msg[COM_MSGSIZE];
   data_t dat;
   file_t file;
   gps_t gps;
   int watchdog, wdt_timeout;
   int usleep_time;
   struct statfs stat;

#ifdef _CYCLE_COUNTS
   cycle_t start_count;
   cycle_t final_count;
#endif
 
   progname = argv[0];
   dbg_fp = stdout;
   err_fp = stdout;

   // turn led 1 on to show startup activity
   led_open();
   led_set_on(1);
   led_set_off(3);

   // mount data file system
   rtn = system("/bin/mount /dev/sda1 /mnt");
   fprintf(stdout, "Data filesystem mounted\n");

   // check disk space
   if((statfs("/mnt", &stat)) < 0 ) {
      fprintf(err_fp, "Failed to stat /mnt\n");
   }
   blksize = (unsigned long)(stat.f_bsize);
   file.bsize = blksize;  // fs block size in bytes
   file.bfree = (unsigned long)(stat.f_bavail);
   file.blocks = (unsigned long)(stat.f_blocks);

   // intialize ADC params
   adc_t *adc; // adc struct
   int adc_fs = 125000;  // sampling freq Hz
   float adc_duration = 4.0; // duration of each adc buffer (seconds)
   int adc_gain = 0;
   int adc_shift = 8; // bit shift
   int adc_nsize = 2; // adc word size

   // COM port params
   strcpy(tty_device, "/dev/ttyBF1");
   strcpy(com.file, "/mnt/com.log"); // com logging

   // initialize data file params
   file.nbufs = 0;
   file.max_nbufs = 10; // max number of data buffer to write per file 
   file.bfree_min = 8000; // 8000*4096 = 32 Mbytes free
   file.format = DAT_FILE_FORMAT;  // default format
   strcpy(file.prefix, "wispr_");
   strcpy(file.path, "/mnt");
   file.log[0] = 0;

   // Watchdog timer
   watchdog = 1;     // watchdog on
   wdt_timeout = 20; // watchdog timer timeout

   max_count = -1;
   usleep_time = 0;

   while ((opt = getopt(argc, argv, "t:s:g:c:n:b:f:l:p:w:v:d:a")) != EOF) {
     switch (opt) {
      case 'f': // adc buffer duration
        adc_fs = atoi(optarg);
        break;
      case 't': // adc buffer duration
        adc_duration = atof(optarg);
        break;
      case 's': // bit shift
        adc_shift = atoi(optarg);
        break;
      case 'g': // gain
        adc_gain = atoi(optarg);
        break;
      case 'c': // number of buffers to loop over
        max_count = atoi(optarg);
        break;
      case 'b': //
        file.max_nbufs = atoi(optarg);
        break;
      case 'l': // log file name
        strncpy(file.log, optarg, 32);
        break;
        strcpy(file.prefix, optarg);
      case 'p': // output file prefix
        break;
      case 'w':
        wdt_timeout = atoi(optarg);
        break;
      case 'v':
        verbose = atoi(optarg);
        break;
      case 'd':
        debug_level = atoi(optarg);
        break;
     }
   }

   // open AD7766 device
   strcpy(adc_device, "/dev/sport0");
   adc = ad7766_open(adc_device, adc_fs);
   if(adc == NULL) {  
      fprintf(err_fp, "Error openning sport0\n"); 
      return(0); 
   }

   // set adc parameters
   adc->gain = adc_gain;
   adc->nsize = adc_nsize;
   adc->shift = adc_shift;

   // allocate adc dma buffers
   if(ad7766_alloc_dma_buffers(adc, adc_duration, adc_nsize) < 0) {
	  fprintf(err_fp, "Error allocating adc dma buffers\n"); 
      goto wispr_close;
   }

   // open gpio lines for pre-amp control
   gpio_export(GPIO_SD); gpio_dir_out(GPIO_SD); // shutdown bit
   gpio_export(GPIO_G1); gpio_dir_out(GPIO_G1); // gain G1 bit
   gpio_export(GPIO_G0); gpio_dir_out(GPIO_G0); // gain G0 bit

   // turn pre-amp ON
   gpio_write(GPIO_SD, 1);

   // set pre-amp gain
   set_adc_gain(adc->gain);

   // open com port
   com.gps = &gps;
   if(com_open (tty_device, B9600) < 0) {
      //perror ("com_open");
      fprintf(err_fp, "%s: error opening tty device\n", progname);
      goto wispr_close;
   }

   // open watchdog
   if(wdt_timeout == 0) watchdog = 0;
   if(watchdog) wdt_open(wdt_timeout);
 
   // start AD7766 clocks
   if(ad7766_start_clock(adc) <= 0) {
      fprintf(err_fp, "Error starting adc clocks\n"); 
      goto wispr_close;
   }

   // start adc dma read cycle using data buffers
   if(ad7766_start_dma(adc) < 0) {
      fprintf(err_fp, "Error starting ad7766\n");
      ad7766_close(adc);
      goto wispr_close;
   }

   // set data struct with current adc values
   dat.nsize = adc->nsize; 
   dat.nsamps = adc->nsamps;
   dat.nchans = adc->nchans;
   dat.fs = adc->fs;
   dat.gain = adc->gain;
   dat.scale = adc->shift;

   // print startup banner
   print_banner(&file, &dat);

   // initial state 
   state = COM_RUN;
   prev_state = COM_RUN;
   mode = 0;
   com.state = state;
   com.mode = mode;

   // turn leds off
   led_set_off(1);
   led_set_off(3);

   // loop over buffers
   count = 0;
   ndtx = 0;
   while(state != COM_EXIT) {

      // keep watchdog alive
      if(watchdog) wdt_keep_alive();

      // check for com actions 
      nrd = com_read_msg (com_msg, COM_MSGSIZE);

      // parse msg, if available
	  // blink led 1 to show com activity
      if (nrd > 0) {
         led_set_on(1);
         state = com_parse_msg (com_msg, nrd, &com);
         mode = com.mode;
         // stop the adc, if running
         if((state == COM_PAUSE) & (prev_state == COM_RUN)) {
            //ad7766_stop(adc);
            fprintf(dbg_fp,"%s: paused\n", progname);
         }
         // start adc, if not already running
         if((state == COM_RUN) & (prev_state != COM_RUN)) {
            fprintf(dbg_fp,"%s: run\n", progname);
            //ad7766_start(adc, data_buffer);
         }
         if(verbose) { 
            sprintf(str,"state=0x%x, mode=0x%x", state, mode);
            print_log(&file, str);
         }
         prev_state = state;
         led_set_off(1);
      }

      // read adc dma data buffer (non-blocking)
      // if a dma buffer is ready, then nrd > 0 and 
      // a non-zero active buffer pointer will be set 
      nrd = ad7766_read_dma(adc, &active_buffer);

      // if paused, continue to read adc, but don't do anything with the data
      if(state == COM_PAUSE) {
         sleep(1);
         continue;
      }
      if(state == COM_EXIT) break; 

      // when a new data bufer is ready, process it, 
      // make sure processing finishes before next buffer is done
      // blink led 3 to show activity
      if(nrd > 0) {
 
         led_set_on(3);

		 // parse adc buffer, convert 32 bits to 16 bits
         ad7766_scale(adc, active_buffer);

         // update active data record
         dat.time.tv_sec = adc->time.tv_sec;
         dat.time.tv_usec = adc->time.tv_nsec/1000;
         dat.scale = adc->shift;
         dat.buffer = (void *)active_buffer;
         dat.status = 0;
         if(adc->clip) dat.status |= DATA_CLIPPED;
         if(adc->dma_buffer.overflow) dat.status |= DATA_OVERRUN;
 
         // write data to file
         nwrt = file_write(&file, &dat);

		 led_set_off(3);

		 count++;

      } 

      // slow things down
      if(usleep_time) usleep(usleep_time);
         
      if((max_count>0) && (count>=max_count)) break; 

   }

wispr_close:

   // turn pre-amp FFO
   gpio_write(GPIO_SD, 0);

   usleep(1000);

   // clean up 
   com_close();
   ad7766_stop_dma(adc);
   ad7766_close(adc);
   gpio_unexport(GPIO_SD);
   gpio_unexport(GPIO_G1);
   gpio_unexport(GPIO_G0);
   wdt_close();
   led_close();

   // unmount data file system 
   //count = 0;
   //while((n != 0) && (count < 4)) {
   //   n = system("/bin/umount /mnt");
   //   fprintf(stdout,"/bin/umount /mnt returned: %d\n",n);
   //   sleep(1);
   //   count++;
   //}

   return 0;
}

int set_adc_gain(int gain) 
{
   if(gain == 0) {
      gpio_write(GPIO_G0, '0'); // pre-amp G0=0
      gpio_write(GPIO_G1, '0'); // pre-amp G1=0
   } else if(gain == 1) {
      gpio_write(GPIO_G0, '1'); // pre-amp G0=0
      gpio_write(GPIO_G1, '0'); // pre-amp G1=0
   } else if(gain == 2) {
      gpio_write(GPIO_G0, '0'); // pre-amp G0=0
      gpio_write(GPIO_G1, '1'); // pre-amp G1=0
   } else if(gain == 3) {
      gpio_write(GPIO_G0, '1'); // pre-amp G0=0
      gpio_write(GPIO_G1, '1'); // pre-amp G1=0
   }
   return(gain); 
}

