/*
* Wave file IO
* ------
* chris jones
*
* The canonical WAVE format starts with the RIFF header:
* Offset  Size Name           Description
*------------------------------------------------------------------------------
* 0       4    ChunkID        Contains the letters "RIFF" in ASCII form
*                             (0x52494646 big-endian form).
* 4       4    ChunkSize      36 + SubChunk2Size, or more precisely:
*                             4 + (8 + SubChunk1Size) + (8 + SubChunk2Size)
*                             This is the size of the rest of the chunk 
*                             following this number.  This is the size of the 
*                             entire file in bytes minus 8 bytes for the
*                             two fields not included in this count:
*                             ChunkID and ChunkSize.
* 8       4    Format          Contains the letters "WAVE"
*                             (0x57415645 big-endian form).
*
* The "WAVE" format consists of two subchunks: "fmt " and "data":
* The "fmt " subchunk describes the sound data's format:
*
* 12      4    Subchunk1ID    Contains the letters "fmt "
*                             (0x666d7420 big-endian form).
* 16      4    Subchunk1Size  16 for PCM.  This is the size of the
*                             rest of the Subchunk which follows this number.
* 20      2    AudioFormat    PCM = 1 (i.e. Linear quantization)
*                             Values other than 1 indicate some 
*                             form of compression.
* 22      2    NumChannels    Mono = 1, Stereo = 2, etc.
* 24      4    SampleRate     8000, 44100, etc.
* 28      4    ByteRate       == SampleRate * NumChannels * BitsPerSample/8
* 32      2    BlockAlign     == NumChannels * BitsPerSample/8
*                             The number of bytes for one sample including
*                             all channels. I wonder what happens when
*                             this number isn't an integer?
* 34      2    BitsPerSample  8 bits = 8, 16 bits = 16, etc.
*         2    ExtraParamSize if PCM, then doesn't exist
*         X    ExtraParams    space for extra parameters

* The "data" subchunk contains the size of the data and the actual sound:
*
* 36      4    Subchunk2ID    Contains the letters "data"
*                             (0x64617461 big-endian form).
* 40      4    Subchunk2Size  == NumSamples * NumChannels * BitsPerSample/8
*                             This is the number of bytes in the data.
*                             You can also think of this as the size
*                             of the read of the subchunk following this 
*                             number.
* 44      *    Data           The actual sound data.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>

#include "wave.h"

size_t bwrite(char *ptr, size_t bsize, size_t nbytes, FILE *stream)
{
  size_t m, n, nblks;
  size_t nwrt = 0;

  // write in blocks, this avoids problems with large buffers
  nblks = nbytes/bsize;
  m = 0;
  for(n = 0; n < nblks; n++) {
    nwrt += fwrite(&ptr[m], 1, bsize, stream);
    m += bsize;
  }
  // write the left over data - if nsamps is not a multiple of bsize
  if(m < nbytes) {
    nwrt += fwrite(&ptr[m], 1, nbytes-m, stream);
  }

  return(nwrt);
}


int wave_read_header(FILE *fp, struct WaveFile *wave)
{
  int nrd;
  nrd = fread(wave->ChunkID,1,4,fp);
  nrd += fread(&(wave->ChunkSize),4,1,fp);
  nrd += fread(wave->Format,1,4,fp);
  nrd += fread(wave->Subchunk1ID,1,4,fp);
  nrd += fread(&(wave->Subchunk1Size),4,1,fp);
  nrd += fread(&(wave->AudioFormat),2,1,fp);
  nrd += fread(&(wave->NumChannels),2,1,fp);
  nrd += fread(&(wave->SampleRate),4,1,fp);
  nrd += fread(&(wave->ByteRate),4,1,fp);
  nrd += fread(&(wave->BlockAlign),2,1,fp);
  nrd += fread(&(wave->BitsPerSample),2,1,fp);
  nrd += fread(&(wave->Subchunk2ID),1,4,fp);
  nrd += fread(&(wave->Subchunk2Size),4,1,fp);
  wave->nsize = (int)(wave->BitsPerSample/8); // sample size in bytes
  wave->nsamps = (int)(wave->Subchunk2Size)/wave->nsize; // num of samps
  return(nrd);
}

int wave_write_header(FILE *fp, struct WaveFile *wave)
{
  int nwrt;
  nwrt = fwrite(wave->ChunkID,1,4,fp);
  nwrt += fwrite(&(wave->ChunkSize),4,1,fp);
  nwrt += fwrite(wave->Format,1,4,fp);
  nwrt += fwrite(wave->Subchunk1ID,1,4,fp);
  nwrt += fwrite(&(wave->Subchunk1Size),4,1,fp);
  nwrt += fwrite(&(wave->AudioFormat),2,1,fp);
  nwrt += fwrite(&(wave->NumChannels),2,1,fp);
  nwrt += fwrite(&(wave->SampleRate),4,1,fp);
  nwrt += fwrite(&(wave->ByteRate),4,1,fp);
  nwrt += fwrite(&(wave->BlockAlign),2,1,fp);
  nwrt += fwrite(&(wave->BitsPerSample),2,1,fp);
  nwrt += fwrite(&(wave->Subchunk2ID),1,4,fp);
  nwrt += fwrite(&(wave->Subchunk2Size),4,1,fp);
  return(nwrt);
}

int wave_read(char *file, struct WaveFile *wave, short *buffer)
{
  int nrd, bufsize, m, n;
  FILE *fp;
  short *tmp;

  // Read the wave file
  fp = fopen(file,"r");
  if(fp <= 0) { 
    fprintf(stderr, "wave_read: error openning %s\n", file);
    return(-1);
  }

  // read header
  nrd = wave_read_header(fp, wave);

  // copy wave data in buffer
  bufsize = 1024;
  tmp = (short *)malloc(wave->nsize*bufsize);
  if(tmp == NULL) {
     fprintf(stdout, "wave_read: data malloc error\n");
     return(-1);
  }
  m = 0;
  nrd = 0;
  while(m < wave->nsamps) {
    nrd += fread(tmp, wave->nsize, bufsize, fp);
    for(n=0; n<bufsize; n++) buffer[m++] = tmp[n];
  }
  free(tmp);

  wave->Data = buffer;

  if(nrd != wave->nsamps) {
    fprintf(stdout, "wave_read: data read error, %d\n", nrd);
  }

  fclose(fp);
  return(nrd);
}

int wave_read_char(char *file,  struct WaveFile *wave, short *buffer, int nsamps, int nsize, int offset)
{
  int m, nrd;
  FILE *fp;

  // Read the wave file
  fp = fopen(file,"r");
  if(fp <= 0) { 
    fprintf(stderr, "wave_read_char: error openning %s\n", file);
    return(-1);
  }

  // read header
  nrd = wave_read_header(fp, wave);

  for(m=0; m < nsamps; m++) {
    buffer[m] = 0;
  }

  if(wave->nsize != nsize) fprintf(stderr, "wave_read_char: nsize problem\n");
  if(wave->nsamps < nsamps) {
    nsamps = wave->nsamps; // num of samps
    fprintf(stderr, "wave_read_char: nsamps = %d\n", nsamps);
  }

  //wave->nsize = nsize; // sample size in bytes

  // read offset words
  for(m=0; m < offset; m++) {
    if(fread(buffer, nsize, 1, fp) <= 0) break;
  }

  // copy wave data in buffer
  m = 0;
  while(m < nsamps) {
    if(fread(&buffer[m++], nsize, 1, fp) <= 0) break;
  }

  if(m != nsamps) {
    fprintf(stdout, "wave_read_char: data read error, %d\n", nrd);
  }

  wave->Data = buffer; // num of samps

  fclose(fp);
  return(m);
}

int wave_write(char *file, struct WaveFile *wave)
{
  int nwrt;
  FILE *fp;
  size_t nsize, nsamps;

  // Write the same file
  fp = fopen(file,"w");
  if(fp <= 0) { 
    fprintf(stderr, "wave_write: error openning %s\n", file);
    return(-1);
  }

  // write header
  nwrt = wave_write_header(fp, wave);

  nsize = (size_t)(wave->BitsPerSample/8); // sample size in bytes
  nsamps = (size_t)(wave->Subchunk2Size/wave->nsize); // num of samps
  if(wave->nsamps != nsamps) { 
    fprintf(stderr, "wave_write: inconsistant nsamps\n");
  }
  // write data
  nwrt = fwrite(wave->Data, wave->nsize, wave->nsamps, fp);
  if(nwrt < nsamps) { 
    fprintf(stderr, "wave_write: error reading data in %s\n", file);
  }
  fclose(fp);
  return(nwrt);
}

int wave_append(char *file, struct WaveFile *wave, char *data, int nsamps, int nsize)
{
  int nrd, nwrt;
  int bsize, nbytes;
  FILE *fp;

  // open for append
  fp = fopen(file,"a+");
  if(fp <= 0) { 
    fprintf(stderr, "wave_append: error openning %s\n", file);
    return(-1);
  }

  // read header
  nrd = wave_read_header(fp, wave);

  // rewrite the updated header
  wave->Subchunk2Size += (int)(nsamps * nsize); // add number of bytes 
  wave->nsamps += nsamps; // add number of samples
  fseek(fp, 0, SEEK_SET);  
  nwrt = wave_write_header(fp, wave);

  // seek to end of file (append to existing data)
  fseek(fp, 0, SEEK_END);  

  // write data in blocks, this avoids problems with large buffers 
  bsize = 4096;
  nbytes = nsize*nsamps;
  nwrt = bwrite(data, bsize, nbytes, fp);
  if(nwrt < nbytes) { 
      fprintf(stderr, "wave_append: error writing data to %s\n", file);
  }

  fclose(fp);
  return(nwrt);
}

int wave_info(char *file, struct WaveFile *wave)
{
  int nrd;
  FILE *fp;
  char str[5];

  // Read the wave file
  fp = fopen(file,"r");
  if(fp <= 0) { 
    fprintf(stderr, "wave_info: error openning %s\n", file);
    return(-1);
  }
  nrd = fread(wave->ChunkID,1,4,fp);
  nrd += fread(&(wave->ChunkSize),4,1,fp);
  nrd += fread(wave->Format,1,4,fp);
  nrd += fread(wave->Subchunk1ID,1,4,fp);
  nrd += fread(&(wave->Subchunk1Size),4,1,fp);
  nrd += fread(&(wave->AudioFormat),2,1,fp);
  nrd += fread(&(wave->NumChannels),2,1,fp);
  nrd += fread(&(wave->SampleRate),4,1,fp);
  nrd += fread(&(wave->ByteRate),4,1,fp);
  nrd += fread(&(wave->BlockAlign),2,1,fp);
  nrd += fread(&(wave->BitsPerSample),2,1,fp);
  nrd += fread(&(wave->Subchunk2ID),1,4,fp);
  nrd += fread(&(wave->Subchunk2Size),4,1,fp);
  fclose(fp);

  wave->nsize = (int)(wave->BitsPerSample/8); // sample size in bytes
  wave->nsamps = (int)(wave->Subchunk2Size)/wave->nsize; // num of samps
  wave->Data = 0;

  fprintf(stdout, "FileName:       %s\n", file); 
  strncpy(str, wave->ChunkID,4); str[4]=0;
  fprintf(stdout, "ChunkID:        %s\n", str); 
  fprintf(stdout, "ChunkSize:      %d\n", wave->ChunkSize);
  strncpy(str, wave->Format,4); str[4]=0;
  fprintf(stdout, "Format:         %s\n", str);
  strncpy(str, wave->Subchunk1ID,4); str[4]=0;
  fprintf(stdout, "Subchunk1ID:    %s\n", str);
  fprintf(stdout, "Subchunk1Size:  %d\n", wave->Subchunk1Size);
  fprintf(stdout, "AudioFormat:    %d\n", wave->AudioFormat);
  fprintf(stdout, "NumChannels:    %d\n", wave->NumChannels);
  fprintf(stdout, "SampleRate:     %d\n", wave->SampleRate);
  fprintf(stdout, "ByteRate:       %d\n", wave->ByteRate);
  fprintf(stdout, "BlockAlign:     %d\n", wave->BlockAlign);
  fprintf(stdout, "BitsPerSample:  %d\n", wave->BitsPerSample);
  strncpy(str, wave->Subchunk2ID,4); str[4]=0;
  fprintf(stdout, "Subchunk2ID:    %s\n", str);
  fprintf(stdout, "Subchunk2Size:  %d\n", wave->Subchunk2Size);
  fprintf(stdout, "nsamps:         %d\n", wave->nsamps);
  fprintf(stdout, "nsize:          %d\n", wave->nsize);

  return(nrd);
}



